from typing import Dict


def keyval_to_dict(key, value, prefix="", separator="__") -> Dict:
    """
    Transform key-value into Mapping.
    """
    parts = key.strip(separator).split(separator, maxsplit=1)
    if len(parts) > 1:
        k, subkey = parts  # unpack split parts for readability
        return {k: keyval_to_dict(subkey, value, prefix=prefix, separator=separator)}
    else:
        return {parts[0]: value}



from ilexconf import Config, from_json

import pytest


def test_copy(settings_json_dict):
    config = Config(settings_json_dict)

    copy = config.copy()

    # Change value in initial object and check it
    config.database.connection.host = "1.2.3.4"
    assert config.database.connection.host == "1.2.3.4"

    # Make sure value in copied object has not changed
    assert copy.database.connection.host == "localhost"



from ilexconf import Config

import pytest


def test_simple_flatten(settings_json_dict):
    config = Config(settings_json_dict)

    flat = config.flatten()
    assert flat == {
        "database.connection.host": "localhost",
        "database.connection.port": 5432,
    }


def test_list_flatten():
    pass



from ilexconf import Config

import pytest


def test_simple_as_table(settings_json_dict):
    config = Config(settings_json_dict)
    headers, rows = config.as_table()

    assert headers == ["Setting", "Value"]
    assert rows == [
        ["database", ""],
        ["  connection", ""],
        ["    host", "localhost"],
        ["    port", "5432"],
    ]


def test_limit_as_table(settings_json_dict):
    config = Config(settings_json_dict)
    headers, rows = config.as_table(limit=3)

    assert rows == [
        ["database", ""],
        ["  connection", ""],
        ["    host", "localhost"],
        ["...", "..."],
    ]



import os
import json

from ilexconf.config import Config
from ilexconf.helpers import keyval_to_dict

from typing import Any, Mapping, Dict, Union, TextIO


def from_env(prefix="", separator="__", lowercase=False):
    """
    Read config from environment variables.
    """
    config = Config()

    for k, v in os.environ.items():
        if prefix and not k.startswith(prefix):
            # If prefix is specified, then ignore the values without it
            print(f"Skipping {k}")
            continue

        # Strip key off of prefix
        prefixless_key = k[len(prefix) :]
        print(f"prefixless {prefixless_key}")
        # Lowercase key if needed
        key = prefixless_key.lower() if lowercase else prefixless_key
        # Convert current key-value pair to Mapping
        d = keyval_to_dict(key, v, prefix=prefix, separator=separator)
        # Merge this Mapping into config
        config.merge(d)

    return config


def to_env(data: Mapping[Any, Any]):
    """
    Set environment variales for child processes invoked using
    methods like os.system(), popen() or fork() and execv().
    """
    config = Config(data)


def from_json(data: Union[str, TextIO], read_from_file: bool = True):
    """
    Read data from JSON file or JSON string.
    """
    if read_from_file:
        if isinstance(data, str):
            json_dict = json.load(open(data, "rt"))
        else:
            json_dict = json.load(data)
    else:
        json_dict = json.loads(str(data))

    config = Config(json_dict)
    return config


def to_json(data: Mapping[Any, Any], path: str):
    """
    Write data to JSON file.
    """
    if isinstance(data, Config):
        data = data.as_dict()
    with open(path, "w") as f:
        json.dump(data, f, indent=2)


def from_ini(
    data: Union[str, TextIO], read_from_file: bool = False, lowercase_keys: bool = False
):
    """
    Read data from INI file or INI string.
    """
    import configparser

    if read_from_file:
        if isinstance(data, str):
            data = open(data, "rt").read()
        else:
            data = data.read()
    data = str(data)
    cfg = configparser.RawConfigParser()
    cfg.read_string(data)
    result = {
        section + "." + k: v
        for section, values in cfg.items()
        for k, v in values.items()
    }



from ilexconf import Config

# from ilexconf.tests.debug import debug

import json
import pytest


NESTED = {
    "a1": {"b1": {"c1": 1, "C2": 2, "c3": 3}, "b2": {"c1": "a", "c2": True, "c3": 1.1}},
    "d1": False,
}

MERGABLE = {"a1": {"b1": {"c3": "val"}}}

MERGED = {
    "a1": {
        "b1": {"c1": 1, "C2": 2, "c3": "val"},
        "b2": {"c1": "a", "c2": True, "c3": 1.1},
    },
    "d1": False,
}

WITH_LISTS = {
    "database": {
        "connections": [
            {"host": "1.2.3.4", "port": 5000},
            {"host": "4.5.6.7", "port": 8080},
        ],
        "params": [["kw1", "kw2", 3], {"key": "value of key"}],
    }
}


def test_empty_config():
    cfg = Config()
    assert cfg == {}
    assert "a0" not in cfg


def test_nested_access():
    cfg = Config(NESTED)
    # Plain access
    assert cfg["a1"]["b1"]["c1"] == 1
    assert cfg["a1.b1.c1"] == 1
    assert cfg.a1.b1.c1 == 1
    # Mixed access
    assert cfg["a1"].b1.C2 == 2
    assert cfg.a1["b1"].c3 == 3
    assert cfg.a1.b2["c3"] == 1.1
    # Direct access
    assert cfg.d1 is False
    assert cfg["d1"] is False


def test_simple_merge():
    cfg = Config(Config({"c1": 1, "C2": 2, "c3": 3}), Config({"c3": "val"}))
    assert cfg.as_dict() == {"c1": 1, "C2": 2, "c3": "val"}


def test_nested_merge():
    cfg = Config(NESTED, MERGABLE)
    # "val" has replaced 3 as value for a1.b1.c3
    assert cfg.a1.b1.c3 == "val"
    # other values are still intact
    assert cfg.a1.b2.c3 == 1.1
    assert cfg.a1.b1.c1 == 1


def test_list():
    cfg = Config(WITH_LISTS)
    print(cfg)
    assert cfg.as_dict() == WITH_LISTS


def test_as_dict():
    cfg = Config(NESTED, MERGABLE)
    d = cfg.as_dict()
    assert d == MERGED


def test_setitem():
    cfg = Config(NESTED, MERGABLE)
    cfg["a1"]["b1"]["c1"] = 6
    assert cfg["a1"]["b1"]["c1"] == 6
    assert cfg.a1.b1.c1 == 6

    empty_cfg = Config()
    empty_cfg["f1"] = False
    assert empty_cfg["f1"] is False
    empty_cfg["f2"]["g3"]["h3"] = 123
    assert empty_cfg["f2.g3.h3"] == 123
    empty_cfg["u1.u4.u6"] = True
    assert empty_cfg["u1"]["u4"]["u6"] is True


def test_setattr():
    cfg = Config(NESTED, MERGABLE)
    cfg.a1.b1.c1 = 8
    assert cfg.a1.b1.c1 == 8
